/* $Id: sdl_v.c 3511 2006-02-01 07:36:15Z tron $ */

#include "../stdafx.h"

#ifdef WITH_SDL

#include "../openttd.h"
#include "../debug.h"
#include "../functions.h"
#include "../gfx.h"
#include "../macros.h"
#include "../sdl_c.h"
#include "../window.h"
#include "../network.h"
#include "../variables.h"
#include "sdl_v.h"
#include <SDL.h>

extern int g_HIDPI_Width;
extern int g_HIDPI_Height;
extern bool hires;

Rect screenrect;
bool screenrotated;
static SDL_Surface *_physical_screen;
static SDL_Surface *_main_screen;
static bool _all_modes;

static SDL_Rect _dirty_rects[MAX_DIRTY_RECTS];
int _num_dirty_rects;

static void SdlVideoMakeDirty(int left, int top, int width, int height)
{
	if (_num_dirty_rects < MAX_DIRTY_RECTS) {
		_dirty_rects[_num_dirty_rects].x = left;
		_dirty_rects[_num_dirty_rects].y = top;
		_dirty_rects[_num_dirty_rects].w = width;
		_dirty_rects[_num_dirty_rects].h = height;
	}
	_num_dirty_rects++;
}

static void UpdatePalette(uint start, uint count)
{
	SDL_Color pal[256];
	uint i;

	for (i = 0; i != count; i++) {
		pal[i].r = _cur_palette[start + i].r;
		pal[i].g = _cur_palette[start + i].g;
		pal[i].b = _cur_palette[start + i].b;
		pal[i].unused = 0;
	}

	SDL_CALL SDL_SetColors(_main_screen, pal, start, count);
}

static void InitPalette(void)
{
	UpdatePalette(0, 256);
}

static void CheckPaletteAnim(void)
{
	if (_pal_last_dirty != -1) {
		UpdatePalette(_pal_first_dirty, _pal_last_dirty - _pal_first_dirty + 1);
		_pal_last_dirty = -1;
	}
}

static void DrawSurfaceToScreen(void)
{
	SDL_Rect sdlrect_src;
	SDL_Rect sdlrect_dst;
#ifdef WITH_DIRTY_RECTS
	int n = _num_dirty_rects;
	if (n != 0) {
		_num_dirty_rects = 0;
		if (n > MAX_DIRTY_RECTS)
			SDL_CALL SDL_UpdateRect(_main_screen, 0, 0, 0, 0);
		else
			SDL_CALL SDL_UpdateRects(_main_screen, n, _dirty_rects);
	}
#else
	sdlrect_dst.x = 0;
	sdlrect_dst.y = 0;
	sdlrect_dst.w = _physical_screen->w;
	sdlrect_dst.h = _physical_screen->h;

	sdlrect_src.x = screenrect.left;
	sdlrect_src.y = screenrect.top;
	sdlrect_src.w = screenrect.right - screenrect.left;
	sdlrect_src.h = screenrect.bottom - screenrect.top;

	SDL_CALL SDL_LowerBlit(_main_screen, &sdlrect_src, _physical_screen, &sdlrect_dst);
	SDL_CALL SDL_UpdateRect(_physical_screen, 0, 0, 0, 0);
#endif
}

static const uint16 default_resolutions[][2] = {
	{ 640,  480}
};

static void GetVideoModes(void)
{
	int i;
	SDL_Rect **modes;

	modes = SDL_CALL SDL_ListModes(NULL, SDL_SWSURFACE + (_fullscreen ? SDL_FULLSCREEN : 0));

	if (modes == NULL)
		error("sdl: no modes available");

	_all_modes = (modes == (void*)-1);

	if (_all_modes) {
		// all modes available, put some default ones here
		memcpy(_resolutions, default_resolutions, sizeof(default_resolutions));
		_num_resolutions = lengthof(default_resolutions);
	} else {
		int n = 0;
		for (i = 0; modes[i]; i++) {
			int w = modes[i]->w;
			int h = modes[i]->h;
			if (IS_INT_INSIDE(w, 640, MAX_SCREEN_WIDTH + 1) &&
					IS_INT_INSIDE(h, 480, MAX_SCREEN_HEIGHT + 1)) {
				int j;
				for (j = 0; j < n; j++) {
					if (_resolutions[j][0] == w && _resolutions[j][1] == h) break;
				}

				if (j == n) {
					_resolutions[j][0] = w;
					_resolutions[j][1] = h;
					if (++n == lengthof(_resolutions)) break;
				}
			}
		}
		_num_resolutions = n;
		SortResolutions(_num_resolutions);
	}
}

static int GetAvailableVideoMode(int *w, int *h)
{
	int i;
	int best;
	uint delta;

	// all modes available?
	if (_all_modes)
		return 1;

	// is the wanted mode among the available modes?
	for (i = 0; i != _num_resolutions; i++) {
		if (*w == _resolutions[i][0] && *h == _resolutions[i][1])
			return 1;
	}

	// use the closest possible resolution
	best = 0;
	delta = abs((_resolutions[0][0] - *w) * (_resolutions[0][1] - *h));
	for (i = 1; i != _num_resolutions; ++i) {
		uint newdelta = abs((_resolutions[i][0] - *w) * (_resolutions[i][1] - *h));
		if (newdelta < delta) {
			best = i;
			delta = newdelta;
		}
	}
	*w = _resolutions[best][0];
	*h = _resolutions[best][1];
	return 2;
}

extern const char _openttd_revision[];

static bool CreateMainSurface(int w, int h)
{
	SDL_Surface *newscreen;
	char caption[50];

	// EDIT GetAvailableVideoMode(&w, &h);

	DEBUG(driver, 1) ("sdl: using mode %dx%d", w, h);

	// DO NOT CHANGE TO HWSURFACE, IT DOES NOT WORK
	newscreen = SDL_CALL SDL_SetVideoMode(320, 240, 8, SDL_SWSURFACE | SDL_HWPALETTE | SDL_FULLSCREEN);
	if (newscreen == NULL)
		return false;

	_physical_screen = newscreen;

	_main_screen = SDL_CALL SDL_CreateRGBSurface(SDL_SWSURFACE, 640, 480, 8, 0, 0, 0, 0);
	
	InitPalette();

	snprintf(caption, sizeof(caption), "OpenTTD %s",_openttd_revision);
	SDL_CALL SDL_WM_SetCaption(caption, caption);
	SDL_CALL SDL_ShowCursor(0);

	GameSizeChanged();

	return true;
}

typedef struct VkMapping {
	uint16 vk_from;
	byte vk_count;
	byte map_to;
} VkMapping;

#define AS(x, z) {x, 0, z}
#define AM(x, y, z, w) {x, y - x, z}

static const VkMapping _vk_mapping[] = {
	// Pageup stuff + up/down
	AM(SDLK_PAGEUP, SDLK_PAGEDOWN, WKC_PAGEUP, WKC_PAGEDOWN),
	AS(SDLK_UP,			WKC_UP),
	AS(SDLK_DOWN,		WKC_DOWN),
	AS(SDLK_LEFT,		WKC_LEFT),
	AS(SDLK_RIGHT,	WKC_RIGHT),

	AS(SDLK_HOME,		WKC_HOME),
	AS(SDLK_END,		WKC_END),

	AS(SDLK_INSERT,	WKC_INSERT),
	AS(SDLK_DELETE,	WKC_DELETE),

	// Map letters & digits
	AM(SDLK_a, SDLK_z, 'A', 'Z'),
	AM(SDLK_0, SDLK_9, '0', '9'),

	AS(SDLK_ESCAPE,	WKC_ESC),
	AS(SDLK_PAUSE,  WKC_PAUSE),
	AS(SDLK_BACKSPACE,	WKC_BACKSPACE),

	AS(SDLK_SPACE,		WKC_SPACE),
	AS(SDLK_RETURN,		WKC_RETURN),
	AS(SDLK_TAB,			WKC_TAB),

	// Function keys
	AM(SDLK_F1, SDLK_F12, WKC_F1, WKC_F12),

	// Numeric part.
	// What is the virtual keycode for numeric enter??
	AM(SDLK_KP0, SDLK_KP9, WKC_NUM_0, WKC_NUM_9),
	AS(SDLK_KP_DIVIDE,		WKC_NUM_DIV),
	AS(SDLK_KP_MULTIPLY,	WKC_NUM_MUL),
	AS(SDLK_KP_MINUS,			WKC_NUM_MINUS),
	AS(SDLK_KP_PLUS,			WKC_NUM_PLUS),
	AS(SDLK_KP_ENTER,			WKC_NUM_ENTER),
	AS(SDLK_KP_PERIOD,		WKC_NUM_DECIMAL)
};

static uint32 ConvertSdlKeyIntoMy(SDL_keysym *sym)
{
	const VkMapping *map;
	uint key = 0;
	for (map = _vk_mapping; map != endof(_vk_mapping); ++map) {
		if ((uint)(sym->sym - map->vk_from) <= map->vk_count) {
			key = sym->sym - map->vk_from + map->map_to;
			break;
		}
	}

	// check scancode for BACKQUOTE key, because we want the key left of "1", not anything else (on non-US keyboards)
#if defined(WIN32) || defined(__OS2__)
	if (sym->scancode == 41) key |= WKC_BACKQUOTE;
#elif defined(__APPLE__)
	if (sym->scancode == 10) key |= WKC_BACKQUOTE;
#elif defined(__MORPHOS__)
	if (sym->scancode == 0)  key |= WKC_BACKQUOTE;  // yes, that key is code '0' under MorphOS :)
#elif defined(__BEOS__)
	if (sym->scancode == 17)  key |= WKC_BACKQUOTE;
#elif defined(__SVR4) && defined(__sun)
	if (sym->scancode == 60) key |= WKC_BACKQUOTE;
	if (sym->scancode == 49) key |= WKC_BACKSPACE;
#elif defined(__sgi__)
	if (sym->scancode == 22) key |= WKC_BACKQUOTE;
#else
	if (sym->scancode == 41) key |= WKC_BACKQUOTE; // Linux console
	if (sym->scancode == 49) key |= WKC_BACKQUOTE;
#endif

	// META are the command keys on mac
	if (sym->mod & KMOD_META) key |= WKC_META;
	if (sym->mod & KMOD_SHIFT) key |= WKC_SHIFT;
	if (sym->mod & KMOD_CTRL) key |= WKC_CTRL;
	if (sym->mod & KMOD_ALT) key |= WKC_ALT;
	// these two lines really help porting hotkey combos. Uncomment to use -- Bjarni
	//printf("scancode character pressed %d\n", sym->scancode);
	//printf("unicode character pressed %d\n", sym->unicode);
	return (key << 16) + sym->unicode;
}

extern void DoExitSave(void);

static int PollEvent(void)
{
//	static bool screenscrolling = false;
	SDL_Event ev;
	memset(&ev, 0, sizeof(ev));

	if (!SDL_CALL SDL_PollEvent(&ev))
		return -2;
			
		
	// jam the viewport position code here
	// check edges of screen for stylus input using rotated coords only if
	// the resolution of the game doesnt match the screen size
	if(/* screenrotated && */
		(/*screenscrolling == true ||*/ 
			( ev.button.state == SDL_PRESSED && ev.button.button == SDL_BUTTON_LEFT ) ) )
	{
		int x, y;
		unsigned short screenchanged = 0;
	
		// set x and y according to screen rotation
		x = ev.motion.x; // EDIT: Legacy GAPI rotate support: g_HIDPI_Height - realy;
		y = ev.motion.y; // EDIT: Legacy GAPI rotate support: realx;
	
		if( y > (g_HIDPI_Height-10) )
		{
			screenrect.top += 60;
			screenchanged = 1;
		}
		if( y < 10 )
		{
		 	screenrect.top -= 60;
		 	screenchanged = 1;
		}
		if( x > (g_HIDPI_Width-10) )
		{
 			screenrect.left += 80;
		 	screenchanged = 1;
		}
		if( x < 10 ) 	
		{
			screenrect.left -= 80;
		 	screenchanged = 1;
		}
		
		if( 1 == screenchanged )
		{
			// check security of screenrect
			if( screenrect.left < 0 ) 				screenrect.left = 0;
			if( screenrect.left > (_screen.width-_cur_resolution[0]) ) 	screenrect.left = _screen.width - _cur_resolution[0];
			if( screenrect.top < 0 ) 				screenrect.top = 0;
			if( screenrect.top > (_screen.height-_cur_resolution[1]) ) 	screenrect.top = _screen.height - _cur_resolution[1];
			
			screenrect.right = screenrect.left + _cur_resolution[0];
			screenrect.bottom = screenrect.top + _cur_resolution[1];
			
			// reposition the toolbar and status bar
			if (_game_mode != GM_MENU) 
			{
				PositionMainToolbar(NULL);
			}
			// make sure window is drawn
			MarkWholeScreenDirty();
			_num_dirty_rects = MAX_DIRTY_RECTS;
			
			_left_button_down = false;		
			_left_button_clicked = false;
			//screenscrolling = true;

			// now reposition the cursor so auto scrolling works
			x = ev.motion.x;
			y = ev.motion.y;
			x += screenrect.left;
			y += screenrect.top;

			if (_cursor.fix_at) {
				int dx = x - _cursor.pos.x;
				int dy = y - _cursor.pos.y;
				if (dx != 0 || dy != 0) 
				{
					_cursor.delta.x += dx;
					_cursor.delta.y += dy;
				}
			} else {
				_cursor.delta.x = x - _cursor.pos.x;
				_cursor.delta.y = y - _cursor.pos.y;
				_cursor.pos.x = x;
				_cursor.pos.y = y;
				_cursor.dirty = true;
			}
		}
		else
		{
		//	screenscrolling = false;
		}
	}

	switch (ev.type) {
	case SDL_MOUSEMOTION:
	{
		int x, y;
	
		// set x and y according to screen rotation
		x = ev.motion.x; // EDIT: Legacy GAPI rotate support: g_HIDPI_Height - realy;
		y = ev.motion.y; // EDIT: Legacy GAPI rotate support: realx;
		x += screenrect.left;
		y += screenrect.top;

		if (_cursor.fix_at) {
			int dx = x - _cursor.pos.x;
			int dy = y - _cursor.pos.y;
			if (dx != 0 || dy != 0) 
			{
				_cursor.delta.x += dx;
				_cursor.delta.y += dy;
			}
		} else {
			_cursor.delta.x = x - _cursor.pos.x;
			_cursor.delta.y = y - _cursor.pos.y;
			_cursor.pos.x = x;
			_cursor.pos.y = y;
			_cursor.dirty = true;
		}
		
		break;
	}

	case SDL_MOUSEBUTTONDOWN:
		if (_rightclick_emulate )
		{
			int numkeys;
			Uint8 *keys;
			keys = SDL_CALL SDL_GetKeyState(&numkeys);
			if( keys[SDLK_HOME] ) // TODO: Make sure this is the correct keycode
			{
				ev.button.button = SDL_BUTTON_RIGHT;
			}
		}
		switch (ev.button.button) {
			case SDL_BUTTON_LEFT:
			{
				int x, y;
				_left_button_down = true;
				
				// set x and y according to screen rotation
				x = ev.motion.x; // EDIT: Legacy GAPI rotate support: g_HIDPI_Height - realy;
				y = ev.motion.y; // EDIT: Legacy GAPI rotate support: realx;
				x += screenrect.left;
				y += screenrect.top;

				if (_cursor.fix_at) {
					int dx = x - _cursor.pos.x;
					int dy = y - _cursor.pos.y;
					if (dx != 0 || dy != 0) {
						_cursor.delta.x += dx;
						_cursor.delta.y += dy;
					}
				} else {
					_cursor.delta.x = x - _cursor.pos.x;
					_cursor.delta.y = y - _cursor.pos.y;
					_cursor.pos.x = x;
					_cursor.pos.y = y;
					_cursor.dirty = true;
				}
				break;
			}
			case SDL_BUTTON_RIGHT:
				_right_button_down = true;
				_right_button_clicked = true;
				break;
			case SDL_BUTTON_WHEELUP:
				_cursor.wheel--;
				break;
			case SDL_BUTTON_WHEELDOWN:
				_cursor.wheel++;
				break;
			default:
				break;
		}
		break;

	case SDL_MOUSEBUTTONUP:
		if (_rightclick_emulate) {
			_right_button_down = false;
			_left_button_down = false;
			_left_button_clicked = false;
		} else if (ev.button.button == SDL_BUTTON_LEFT) {
			_left_button_down = false;		
			_left_button_clicked = false;
		} else if (ev.button.button == SDL_BUTTON_RIGHT) {
			_right_button_down = false;
		}
		break;

	case SDL_QUIT:
		// do not ask to quit on the main screen
		if (_game_mode != GM_MENU) {
			if (_patches.autosave_on_exit) {
				DoExitSave();
				return 0;
			} else
				AskExitGame();
		} else
			return 0;
		break;

	case SDL_KEYDOWN: /* Toggle full-screen on ALT + ENTER/F */
		if ((ev.key.keysym.mod & (KMOD_ALT | KMOD_META)) &&
				(ev.key.keysym.sym == SDLK_RETURN || ev.key.keysym.sym == SDLK_f)) {
			ToggleFullScreen(!_fullscreen);
		} else
			_pressed_key = ConvertSdlKeyIntoMy(&ev.key.keysym);

		break;

	case SDL_VIDEORESIZE: {
		int w = clamp(ev.resize.w, 64, MAX_SCREEN_WIDTH);
		int h = clamp(ev.resize.h, 64, MAX_SCREEN_HEIGHT);
		ChangeResInGame(w, h);
		break;
	}
	}
	return -1;
}

static void SdlVideoFullScreen(bool full_screen)
{
#ifdef _WIN32_WCE
	HWND hwndTaskbar = FindWindow(_T("HHTaskbar"), NULL);
	if (NULL == hwndTaskbar)
	{
		// On Smartphones, the taskbar window is renamed to "Tray"
		hwndTaskbar = FindWindow(_T("Tray"), NULL);
	}

	if (full_screen)
	{
/*
		HKEY hKey=0;
		DWORD dwValue = 0;
		unsigned long lSize = sizeof( DWORD );
		DWORD dwType = REG_DWORD;
		HWND hWndTemp;
		RegOpenKeyEx( HKEY_LOCAL_MACHINE, TEXT("\\software\\microsoft\\shell"), 0, KEY_ALL_ACCESS, &hKey );
		RegQueryValueEx( hKey, TEXT("TBOpt"), 0, &dwType, (BYTE*)&dwValue, &lSize );
		dwValue |= 8;	// set bit to hide taskbar
		RegSetValueEx( hKey, TEXT("TBOpt"), 0, REG_DWORD, (BYTE*)&dwValue, lSize );
		hWndTemp = FindWindow( TEXT("HHTaskBar"), NULL );
		SendMessage(hWndTemp, WM_COMMAND, 0x03EA, 0 );
		SetForegroundWindow(hWnd);
		if (FindWindow(TEXT("HHTaskBar"), NULL)) 
		{ // is it valid for HPC ?
			ShowWindow(FindWindow(TEXT("HHTaskBar"),NULL),SW_HIDE);
		}
*/
		if (hwndTaskbar)
		{
			// Hide the task bar
			ShowWindow(hwndTaskbar, SW_HIDE);
		}
	}
	else
	{
		if (hwndTaskbar)
		{
			// Show the task bar
			ShowWindow(hwndTaskbar, SW_SHOW);
		}
	}
#endif

	_fullscreen = full_screen;
}

static const char *SdlVideoStart(const char * const *parm)
{
	char buf[30];

	const char *s = SdlOpen(SDL_INIT_VIDEO);
	if (s != NULL) 
	{
//		MessageBox(NULL, TEXT("SDL CANNOT INIT VIDEO"), TEXT("SDL CANNOT INIT VIDEO"), MB_ICONEXCLAMATION );
		return s;
	}

	SDL_CALL SDL_VideoDriverName(buf, 30);
	DEBUG(driver, 1) ("sdl: using driver '%s'", buf);

	//GetVideoModes();
	_screen.width = 640	;
	_screen.height = 480;
	_screen.pitch = 640;

	// autodetect landscape or portrait mode
	if( g_HIDPI_Width >= g_HIDPI_Height ) screenrotated = false;
	else screenrotated = true;

	// now set to full screen
	if( screenrotated )
	{
		_cur_resolution[1] = g_HIDPI_Width;
		_cur_resolution[0] = g_HIDPI_Height;
	}
	else
	{
		_cur_resolution[0] = g_HIDPI_Width;
		_cur_resolution[1] = g_HIDPI_Height;
	}

	screenrect.left = (_screen.width / 2) - (_cur_resolution[0] / 2); // center
	/*if( hires || !screenrotated ) screenrect.top = 0;
	else */screenrect.top = 70;
	screenrect.right = screenrect.left + _cur_resolution[0];
	screenrect.bottom = screenrect.top + _cur_resolution[1];

	// boundry check
	if( screenrect.right > _screen.width ) screenrect.right = _screen.width;
	if( screenrect.bottom > _screen.height ) screenrect.bottom = _screen.height;

	_cursor.pos.x = screenrect.left + 1;
	_cursor.pos.y = screenrect.top;

	CreateMainSurface(_cur_resolution[0], _cur_resolution[1]);
	MarkWholeScreenDirty();
	
	// set to full screen
	SdlVideoFullScreen(true);

	SDL_CALL SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_CALL SDL_EnableUNICODE(1);
	return NULL;
}

static void SdlVideoStop(void)
{
	SdlClose(SDL_INIT_VIDEO);
	// EDIT Now get back the start menu
	SdlVideoFullScreen(false);
}

static void SdlVideoMainLoop(void)
{
	uint32 next_tick = SDL_CALL SDL_GetTicks() + 30;
	uint32 cur_ticks;
	uint32 pal_tick = 0;
	int i;
	uint32 mod;
	int numkeys;
	Uint8 *keys;

	for (;;) {
		InteractiveRandom(); // randomness

		while ((i = PollEvent()) == -1) {}
		if (i >= 0) return;

		if (_exit_game) return;

		mod = SDL_CALL SDL_GetModState();
		keys = SDL_CALL SDL_GetKeyState(&numkeys);
#if defined(_DEBUG)
		if (_shift_pressed)
#else
		if (keys[SDLK_TAB])
#endif
		{
			if (!_networking && _game_mode != GM_MENU) _fast_forward |= 2;
		} else if (_fast_forward & 2) {
			_fast_forward = 0;
		}

		cur_ticks = SDL_CALL SDL_GetTicks();
		if ((_fast_forward && !_pause) || cur_ticks > next_tick)
			next_tick = cur_ticks;

		if (cur_ticks == next_tick) {
			next_tick += 30;

			_ctrl_pressed = !!(mod & (KMOD_LCTRL | KMOD_RCTRL));
			_shift_pressed = !!(mod & (KMOD_LSHIFT | KMOD_RSHIFT));
#ifdef _DEBUG
			_dbg_screen_rect = !!(mod & KMOD_CAPS);
#endif

			// determine which directional keys are down
			if( screenrotated )
			{
				_dirkeys =
					(keys[SDLK_LEFT]  ? 2 : 0) |
					(keys[SDLK_UP]    ? 4 : 0) |
					(keys[SDLK_RIGHT] ? 8 : 0) |
					(keys[SDLK_DOWN]  ? 1 : 0);
			}
			else
			{
				_dirkeys =
					(keys[SDLK_LEFT]  ? 1 : 0) |
					(keys[SDLK_UP]    ? 2 : 0) |
					(keys[SDLK_RIGHT] ? 4 : 0) |
					(keys[SDLK_DOWN]  ? 8 : 0);
			}
			
			GameLoop();

			_screen.dst_ptr = _main_screen->pixels;
			UpdateWindows();
			if (++pal_tick > 4) {
				CheckPaletteAnim();
				pal_tick = 1;
			}
			DrawSurfaceToScreen();
		} else {
			SDL_CALL SDL_Delay(1);
			_screen.dst_ptr = _main_screen->pixels;
			DrawTextMessage();
			DrawMouseCursor();
			DrawSurfaceToScreen();
		}
	}
}

static bool SdlVideoChangeRes(int w, int h)
{
	return CreateMainSurface(w, h);
}


const HalVideoDriver _sdl_video_driver = {
	SdlVideoStart,
	SdlVideoStop,
	SdlVideoMakeDirty,
	SdlVideoMainLoop,
	SdlVideoChangeRes,
	SdlVideoFullScreen,
};

#endif
